@@ -1,4 +1,3 @@
-# Copyright (c) 2015. Mount Sinai School of Medicine
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,135 +11,47 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-# required so that 'import vcf' gets the global PyVCF package,
-# rather than our local vcf module
-from __future__ import absolute_import
 
+from __future__ import absolute_import, print_function, division
+import os
 import requests
 import zlib
-import collections
+import urllib
+import logging
+from collections import OrderedDict
+from warnings import warn
 
-try:
-    from urlparse import urlparse  # Python 2
-except ImportError:
-    from urllib.parse import urlparse  # Python 3
-
-from pyensembl import genome_for_reference_name
 import pandas
-import typechecks
-import vcf  # PyVCF
+from typechecks import require_string
+import vcf as pyvcf
 
-from .reference_name import infer_reference_name
-from .variant import Variant
+from .reference import (
+    infer_genome,
+    ensembl_to_ucsc_reference_names
+)
+from .variant import Variant, variant_ascending_position_sort_key
 from .variant_collection import VariantCollection
 
+
+logger = logging.getLogger(__name__)
+
+
 def load_vcf(
         path,
         genome=None,
-        only_passing=True,
-        reference_name=None,
         reference_vcf_key="reference",
-        allow_extended_nucleotides=False,
-        max_variants=None):
-    """
-    Load reference name and Variant objects from the given VCF filename.
-
-    This uses PyVCF to parse the file. It is slower than the pandas
-    implementation used in `load_vcf_fast`, but is better tested and more
-    robust.
-
-    Parameters
-    ----------
-
-    path : str or vcf.Reader
-        Path or URL to VCF (*.vcf) or compressed VCF (*.vcf.gz). Supported URL
-        schemes are "file", "http", "https", and "ftp". Can also be a pyvcf
-        Reader instance.
-
-    genome : pyensembl.Genome, optional
-        Optionally pass in a PyEnsembl Genome or EnsemblRelease object
-        to exactly specify the annotation data source
-
-    only_passing : boolean, optional
-        If true, any entries whose FILTER field is not one of "." or "PASS" is
-        dropped.
-
-    reference_name : str, optional
-        Name of reference genome against which variants from VCF were aligned.
-        If specified, then `reference_vcf_key` is ignored.
-
-    reference_vcf_key : str, optional
-        Name of metadata field which contains path to reference FASTA
-        file (default = 'reference')
-
-    allow_extended_nucleotides : boolean, default False
-        Allow characters other that A,C,T,G in the ref and alt strings.
-
-    max_variants : int, optional
-        If specified, return only the first max_variants variants.
-    """
-
-    variants = []
-    metadata = {}
-    handle = PyVCFReaderFromPathOrURL(path)
-    try:
-        if not genome:
-            genome = make_pyensembl_genome_object(
-                handle.vcf_reader,
-                reference_name,
-                reference_vcf_key)
-
-        for record in handle.vcf_reader:
-            if only_passing and record.FILTER and record.FILTER != "PASS":
-                continue
-            for (alt_num, alt) in enumerate(record.ALT):
-                # We ignore "no-call" variants, i.e. those where X.ALT = [None]
-                if not alt:
-                    continue
-                variant = Variant(
-                    contig=record.CHROM,
-                    start=record.POS,
-                    ref=record.REF,
-                    alt=alt.sequence,
-                    ensembl=genome,
-                    allow_extended_nucleotides=allow_extended_nucleotides)
-                variants.append(variant)
-                metadata[variant] = {
-                    "id": record.ID,
-                    "qual": record.QUAL,
-                    "filter": record.FILTER,
-                    "alt_allele_index": alt_num,
-                    "info": dict(record.INFO),
-                }
-                if max_variants and len(variants) > max_variants:
-                    raise StopIteration
-    except StopIteration:
-        pass
-    finally:
-        handle.close()
-
-    return VariantCollection(
-        variants=variants, path=handle.path, metadata=metadata)
-
-def load_vcf_fast(
-        path,
-        genome=None,
         only_passing=True,
-        reference_name=None,
-        reference_vcf_key="reference",
         allow_extended_nucleotides=False,
         include_info=True,
-        chunk_size=10**5,
-        max_variants=None):
+        chunk_size=10 ** 5,
+        max_variants=None,
+        sort_key=variant_ascending_position_sort_key,
+        distinct=True,
+        normalize_contig_names=True,
+        convert_ucsc_contig_names=None):
     """
     Load reference name and Variant objects from the given VCF filename.
 
-    This is an experimental faster implementation of `load_vcf`. It is
-    typically about 2X faster, and with `include_info=False`, about 4X faster.
-    If most of the records in the VCF have failed filters (and
-    only_passing=True), this function can be orders of magnitude faster than
-    `load_vcf`.
-
     Currently only local files are supported by this function (no http). If you
     call this on an HTTP URL, it will fall back to `load_vcf`.
 
@@ -150,37 +61,52 @@ def load_vcf_fast(
     path : str
         Path to VCF (*.vcf) or compressed VCF (*.vcf.gz).
 
-    genome : Genome, optional
-        Optionally pass in a PyEnsembl Genome or EnsemblRelease object
-        to exactly specify the annotation data source
-
-    only_passing : boolean, optional
-        If true, any entries whose FILTER field is not one of "." or "PASS" is
-        dropped.
-
-    reference_name : str, optional
-        Name of reference genome against which variants from VCF were aligned.
-        If specified, then `reference_vcf_key` is ignored.
+    genome : {pyensembl.Genome, reference name, Ensembl version int}, optional
+        Optionally pass in a PyEnsembl Genome object, name of reference, or
+        PyEnsembl release version to specify the reference associated with a
+        VCF (otherwise infer reference from VCF using reference_vcf_key)
 
     reference_vcf_key : str, optional
         Name of metadata field which contains path to reference FASTA
         file (default = 'reference')
 
-    allow_extended_nucleotides : boolean, default False
+    only_passing : bool, optional
+        If true, any entries whose FILTER field is not one of "." or "PASS" is
+        dropped.
+
+    allow_extended_nucleotides : bool, default False
         Allow characters other that A,C,T,G in the ref and alt strings.
 
-    include_info : boolean, default True
-        Whether to parse the info column. If you don't need that column, set to
-        False for faster parsing.
+    include_info : bool, default True
+        Whether to parse the INFO and per-sample columns. If you don't need
+        these, set to False for faster parsing.
 
     chunk_size: int, optional
         Number of records to load in memory at once.
 
     max_variants : int, optional
         If specified, return only the first max_variants variants.
+
+    sort_key : fn
+        Function which maps each element to a sorting criterion.
+        Set to None to not to sort the variants.
+
+    distinct : bool, default True
+        Don't keep repeated variants
+
+    normalize_contig_names : bool, default True
+        By default contig names will be normalized by converting integers
+        to strings (e.g. 1 -> "1"), and converting any letters after "chr"
+        to uppercase (e.g. "chrx" -> "chrX"). If you don't want
+        this behavior then pass normalize_contig_names=False.
+
+    convert_ucsc_contig_names : bool
+        Convert chromosome names from hg19 (e.g. "chr1") to equivalent names
+        for GRCh37 (e.g. "1"). By default this is set to True if the genome
+        of the VCF is a UCSC reference and otherwise set to False.
     """
 
-    typechecks.require_string(path, "Path or URL to VCF")
+    require_string(path, "Path or URL to VCF")
     parsed_path = parse_url_or_path(path)
 
     if parsed_path.scheme and parsed_path.scheme.lower() != "file":
@@ -189,16 +115,34 @@ def load_vcf_fast(
         # implementation of read_table (with engine="python") helps with some
         # issues but introduces a new set of problems (e.g. the dtype parameter
         # is not accepted). For these reasons, we're currently not attempting
-        # to load VCFs over HTTP with pandas, and fall back to the pyvcf
-        # implementation here.
-        return load_vcf(
-            path,
-            genome=genome,
-            only_passing=only_passing,
-            reference_name=reference_name,
-            reference_vcf_key=reference_vcf_key,
-            allow_extended_nucleotides=allow_extended_nucleotides,
-            max_variants=max_variants)
+        # to load VCFs over HTTP with pandas directly, and instead download it
+        # to a temporary file and open that.
+
+        (filename, headers) = urllib.request.urlretrieve(path)
+        try:
+            # The downloaded file has no file extension, which confuses pyvcf
+            # for gziped files in Python 3. We rename it to have the correct
+            # file extension.
+            new_filename = "%s.%s" % (
+                filename, parsed_path.path.split(".")[-1])
+            os.rename(filename, new_filename)
+            filename = new_filename
+            return load_vcf(
+                filename,
+                genome=genome,
+                reference_vcf_key=reference_vcf_key,
+                only_passing=only_passing,
+                allow_extended_nucleotides=allow_extended_nucleotides,
+                include_info=include_info,
+                chunk_size=chunk_size,
+                max_variants=max_variants,
+                sort_key=sort_key,
+                distinct=distinct,
+                normalize_contig_names=normalize_contig_names,
+                convert_ucsc_contig_names=convert_ucsc_contig_names)
+        finally:
+            logger.info("Removing temporary file: %s", filename)
+            os.unlink(filename)
 
     # Loading a local file.
     # The file will be opened twice: first to parse the header with pyvcf, then
@@ -209,30 +153,105 @@ def load_vcf_fast(
     handle = PyVCFReaderFromPathOrURL(path)
     handle.close()
 
-    if not genome:
-        genome = make_pyensembl_genome_object(
-            handle.vcf_reader,
-            reference_name,
-            reference_vcf_key)
+    ####
+    # The following code looks a bit crazy because it's motivated by the
+    # desired to preserve UCSC reference names even though the Variant
+    # objects we're creating will convert them to EnsemblRelease genomes
+    # with different reference names.
+    #
+    # For example, if a VCF is aligned against 'hg19' then we want to create a
+    # variant which has 'hg19' as its genome argument, so that serialization
+    # back to VCF will put the correct reference genome in the generated
+    # header.
+    if genome is None:
+        vcf_reader = handle.vcf_reader
+        if reference_vcf_key not in vcf_reader.metadata:
+            raise ValueError("Unable to infer reference genome for %s" % (
+                vcf_reader.filename,))
+        genome = vcf_reader.metadata[reference_vcf_key]
+
+    genome, genome_was_ucsc = infer_genome(genome)
+    if genome_was_ucsc:
+        genome = ensembl_to_ucsc_reference_names[genome.reference_name]
+
+    if convert_ucsc_contig_names is None:
+        convert_ucsc_contig_names = genome_was_ucsc
 
     df_iterator = read_vcf_into_dataframe(
-        path, include_info=include_info, chunk_size=chunk_size)
+        path,
+        include_info=include_info,
+        sample_names=handle.vcf_reader.samples if include_info else None,
+        chunk_size=chunk_size)
+
+    if include_info:
+        def sample_info_parser(unparsed_sample_info_strings, format_string):
+            """
+            Given a format string like "GT:AD:ADP:DP:FS"
+            and a list of sample info strings where each entry is like
+            "0/1:3,22:T=3,G=22:25:33", return a dict that maps:
+            sample name -> field name -> value. Uses pyvcf to parse the fields.
+            """
+            return pyvcf_calls_to_sample_info_list(
+                handle.vcf_reader._parse_samples(
+                    unparsed_sample_info_strings, format_string, None))
+    else:
+        sample_info_parser = None
+
+    variant_kwargs = {
+        'genome': genome,
+        'allow_extended_nucleotides': allow_extended_nucleotides,
+        'normalize_contig_names': normalize_contig_names,
+        'convert_ucsc_contig_names': convert_ucsc_contig_names,
+    }
+
+    variant_collection_kwargs = {
+        'sort_key': sort_key,
+        'distinct': distinct
+    }
+
+    # TODO: drop chrMT variants from hg19 and warn user about it
 
     return dataframes_to_variant_collection(
         df_iterator,
+        source_path=path,
         info_parser=handle.vcf_reader._parse_info if include_info else None,
         only_passing=only_passing,
         max_variants=max_variants,
-        variant_kwargs={
-            'ensembl': genome,
-            'allow_extended_nucleotides': allow_extended_nucleotides},
-        variant_collection_kwargs={"path": path})
+        sample_names=handle.vcf_reader.samples if include_info else None,
+        sample_info_parser=sample_info_parser,
+        variant_kwargs=variant_kwargs,
+        variant_collection_kwargs=variant_collection_kwargs)
+
+
+
+def load_vcf_fast(*args, **kwargs):
+    """
+    Same as load_vcf, keeping this name for backwards compatibility.
+    """
+    warn(
+        "load_vcf_fast is deprecated and has been renamed to load_vcf",
+        DeprecationWarning)
+    return load_vcf(*args, **kwargs)
+
+
+def pyvcf_calls_to_sample_info_list(calls):
+    """
+    Given pyvcf.model._Call instances, return a dict mapping each sample
+    name to its per-sample info:
+        sample name -> field -> value
+    """
+    return OrderedDict(
+        (call.sample, call.data._asdict()) for call in calls)
+
 
 def dataframes_to_variant_collection(
         dataframes,
+        source_path,
         info_parser=None,
         only_passing=True,
         max_variants=None,
+        sample_names=None,
+        sample_info_parser=None,
         variant_kwargs={},
         variant_collection_kwargs={}):
     """
@@ -250,6 +269,9 @@ def dataframes_to_variant_collection(
         and 'INFO' if `info_parser` is not Null. Columns must be in this
         order.
 
+    source_path : str
+        Path of VCF file from which DataFrame chunks were generated.
+
     info_parser : string -> object, optional
         Callable to parse INFO strings.
 
@@ -260,6 +282,14 @@ def dataframes_to_variant_collection(
     max_variants : int, optional
         If specified, return only the first max_variants variants.
 
+    sample_names : list of strings, optional
+        Sample names. The final columns of the dataframe should match these.
+        If specified, the per-sample info columns will be parsed. You must
+        also specify sample_info_parser.
+
+    sample_info_parser : string list * string -> dict, optional
+        Callable to parse per-sample info columns.
+
     variant_kwargs : dict, optional
         Additional keyword paramters to pass to Variant.__init__
 
@@ -271,6 +301,13 @@ def dataframes_to_variant_collection(
         ["CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER"] +
         (["INFO"] if info_parser else []))
 
+    if info_parser and sample_names:
+        if sample_info_parser is None:
+            raise TypeError(
+                "Must specify sample_info_parser if specifying sample_names")
+        expected_columns.append("FORMAT")
+        expected_columns.extend(sample_names)
+
     variants = []
     metadata = {}
     try:
@@ -293,12 +330,18 @@ def dataframes_to_variant_collection(
                     id_ = None
                 qual = float(qual) if qual != "." else None
                 alt_num = 0
-                info = None
+                info = sample_info = None
                 for alt in alts.split(","):
                     if alt != ".":
                         if info_parser is not None and info is None:
-                            # tpl[-1] is info.
-                            info = info_parser(tpl[-1])
+                            info = info_parser(tpl[8])  # INFO column
+                            if sample_names:
+                                # Sample name -> field -> value dict.
+                                sample_info = sample_info_parser(
+                                    list(tpl[10:]),  # sample info columns
+                                    tpl[9],    # FORMAT column
+                                )
+
                         variant = Variant(
                             chrom,
                             int(pos),  # want a Python int not numpy.int64
@@ -311,6 +354,7 @@ def dataframes_to_variant_collection(
                             'qual': qual,
                             'filter': flter,
                             'info': info,
+                            'sample_info': sample_info,
                             'alt_allele_index': alt_num,
                         }
                         if max_variants and len(variants) > max_variants:
@@ -320,9 +364,16 @@ def dataframes_to_variant_collection(
         pass
 
     return VariantCollection(
-        variants=variants, metadata=metadata, **variant_collection_kwargs)
+        variants=variants,
+        source_to_metadata_dict={source_path: metadata},
+        **variant_collection_kwargs)
 
-def read_vcf_into_dataframe(path, include_info=False, chunk_size=None):
+
+def read_vcf_into_dataframe(
+        path,
+        include_info=False,
+        sample_names=None,
+        chunk_size=None):
     """
     Load the data of a VCF into a pandas dataframe. All headers are ignored.
 
@@ -335,6 +386,11 @@ def read_vcf_into_dataframe(path, include_info=False, chunk_size=None):
         If true, the INFO field is not parsed, but is included as a string in
         the resulting data frame. If false, the INFO field is omitted.
 
+    sample_names: string list, optional
+        Sample names. The final columns of the dataframe should match these.
+        If specified (and include_info is also specified), the FORMAT and
+        per-sample info columns will be included in the result dataframe.
+
     chunk_size : int, optional
         If buffering is desired, the number of rows per chunk.
 
@@ -344,7 +400,7 @@ def read_vcf_into_dataframe(path, include_info=False, chunk_size=None):
     VCF file. Otherwise, an iterable of dataframes, each with chunk_size rows.
 
     """
-    vcf_field_types = collections.OrderedDict()
+    vcf_field_types = OrderedDict()
     vcf_field_types['CHROM'] = str
     vcf_field_types['POS'] = int
     vcf_field_types['ID'] = str
@@ -354,6 +410,10 @@ def read_vcf_into_dataframe(path, include_info=False, chunk_size=None):
     vcf_field_types['FILTER'] = str
     if include_info:
         vcf_field_types['INFO'] = str
+        if sample_names:
+            vcf_field_types['FORMAT'] = str
+            for name in sample_names:
+                vcf_field_types[name] = str
 
     parsed_path = parse_url_or_path(path)
     if not parsed_path.scheme or parsed_path.scheme.lower() == "file":
@@ -367,8 +427,9 @@ def read_vcf_into_dataframe(path, include_info=False, chunk_size=None):
     elif path.endswith(".bz2"):
         compression = "bz2"
 
-    reader = pandas.read_table(
+    reader = pandas.read_csv(
         path,
+        sep="\t",
         compression=compression,
         comment="#",
         chunksize=chunk_size,
@@ -377,6 +438,7 @@ def read_vcf_into_dataframe(path, include_info=False, chunk_size=None):
         usecols=range(len(vcf_field_types)))
     return reader
 
+
 class PyVCFReaderFromPathOrURL(object):
     """
     Thin wrapper over a PyVCF Reader object that supports loading over URLs,
@@ -402,14 +464,16 @@ class PyVCFReaderFromPathOrURL(object):
         self.vcf_reader = None  # vcf_reader. Will always be set.
         self._to_close = None  # object to call close() on when we're done.
 
-        if isinstance(path, vcf.Reader):
+        if isinstance(path, pyvcf.Reader):
             self.vcf_reader = path
         else:
-            typechecks.require_string(path, "Path or URL to VCF")
+            require_string(path, "Path or URL to VCF")
             self.path = path
             parsed_path = parse_url_or_path(path)
             if not parsed_path.scheme or parsed_path.scheme.lower() == 'file':
-                self.vcf_reader = vcf.Reader(filename=parsed_path.path)
+                self.vcf_reader = pyvcf.Reader(
+                    filename=parsed_path.path,
+                    strict_whitespace=True)
             elif parsed_path.scheme.lower() in ("http", "https", "ftp"):
                 self._to_close = response = requests.get(path, stream=True)
                 response.raise_for_status()  # raise error on 404, etc.
@@ -418,7 +482,10 @@ class PyVCFReaderFromPathOrURL(object):
                         response.iter_content())
                 else:
                     lines = response.iter_lines(decode_unicode=True)
-                self.vcf_reader = vcf.Reader(fsock=lines, compressed=False)
+                self.vcf_reader = pyvcf.Reader(
+                    fsock=lines,
+                    compressed=False,
+                    strict_whitespace=True)
             else:
                 raise ValueError("Unsupported scheme: %s" % parsed_path.scheme)
 
@@ -426,6 +493,7 @@ class PyVCFReaderFromPathOrURL(object):
         if self._to_close is not None:
             self._to_close.close()
 
+
 def stream_gzip_decompress_lines(stream):
     """
     Uncompress a gzip stream into lines of text.
@@ -449,24 +517,7 @@ def stream_gzip_decompress_lines(stream):
                 yield line
     yield previous
 
-def make_pyensembl_genome_object(
-        vcf_reader,
-        reference_name,
-        reference_vcf_key):
-    """
-    Helper function to make a pyensembl.Genome instance.
-    """
 
-    if reference_name:
-        # normalize the reference name in case it's in a weird format
-        reference_name = infer_reference_name(reference_name)
-    elif reference_vcf_key not in vcf_reader.metadata:
-        raise ValueError("Unable to infer reference genome for %s" % (
-            vcf_reader.filename,))
-    else:
-        reference_path = vcf_reader.metadata[reference_vcf_key]
-        reference_name = infer_reference_name(reference_path)
-    return genome_for_reference_name(reference_name)
 
 def parse_url_or_path(s):
     # urlparse will parse paths with two leading slashes (e.g. "//foo")
@@ -474,4 +525,4 @@ def parse_url_or_path(s):
     # slash.
     if s.startswith("//"):
         s = "/" + s.lstrip("/")
-    return urlparse(s)
+    return urllib.parse.urlparse(s)
